#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>

// gcc -Wall -W -o bank_conflicts bank_conflicts.c

#define LOCAL_SZ      (48 * 1024)                          // available local memory: 48k
#define KEYS_VALUES   256                                  // 8 bit
#define KEYS_GROUPS   4                                    // we have 4 lookup tables
#define KEY_SZ        4                                    // of type u32
#define SBOX_SIZE     (KEYS_GROUPS * KEYS_VALUES * KEY_SZ) // 4096 byte
#define BANKS         32                                   // fixed on NV and AMD
#define BANK_CHUNK_SZ 4                                    // fixed on NV and AMD
#define CHUNKS        (LOCAL_SZ / BANK_CHUNK_SZ)           // 12288
#define THREADS       (LOCAL_SZ / SBOX_SIZE)               // 12
#define LOOKUP_GAP    THREADS                              // can be 12 if THREADS is 11 for 48k-1 case for OpenCL
//#define THREADS       11
//#define LOOKUP_GAP    12

typedef struct
{
  int t; // thread
  int c; // column
  int k; // key

} entry_t;

typedef uint8_t  u8;
typedef uint16_t u16;

int make_idx (const int lid, const int col, const int key)
{
  // linear
  // 32k              :  7 conflicts
  // 48k (12 threads) : 11 conflicts
  // 48k (11 threads) : 10 conflicts
  // 64k              : 15 conflicts
  // key extra work   :  0 instructions
  //return (lid * KEYS_VALUES * KEYS_GROUPS) + (col * KEYS_VALUES) + key;

  // 32k              :    0 conflicts
  // 48k (12 threads) :    2 conflicts
  // 48k (11 threads) : 1.75 conflicts
  // 64k              :    0 conflicts
  // key extra work   :    2 instructions
  //return (key * KEYS_GROUPS * LOOKUP_GAP) + (col * LOOKUP_GAP) + lid;

  // 32k              :    0 conflicts
  // 48k (12 threads) :    2 conflicts
  // 48k (11 threads) : 1.75 conflicts
  // 64k              :    0 conflicts
  // key extra work   :    1 instructions
  return (col * LOOKUP_GAP * KEYS_VALUES) + (key * LOOKUP_GAP) + lid;
}

int main ()
{
  entry_t m[CHUNKS];

  for (int i = 0; i < CHUNKS; i++)
  {
    m[i].t = -1;
    m[i].c = -1;
    m[i].k = -1;
  }

  // apply the pattern

  for (int t = 0; t < THREADS; t++)
  {
    for (int c = 0; c < KEYS_GROUPS; c++)
    {
      for (int k = 0; k < KEYS_VALUES; k++)
      {
        const int idx = make_idx (t, c, k);

        m[idx].t = t;
        m[idx].c = c;
        m[idx].k = k;
      }
    }
  }

  // visualize mapping
  // 1. pos = 1 byte,  thread
  // 2. pos = 1 byte,  sbox group
  // 3. pos = 2 bytes, sbox key

  int banks_to_threads[BANKS][THREADS];

  for (int row = 0; row < CHUNKS; row += BANKS)
  {
    if ((row % (CHUNKS / BANK_CHUNK_SZ)) == 0)
    {
      printf ("byte offset: %d\n", row * BANK_CHUNK_SZ);
    }

    for (int col = 0; col < BANKS; col++)
    {
      int idx = row + col;

      printf ("%01x%01x%02x ", m[idx].t & 0xf, m[idx].c & 0xf, m[idx].k & 0xff);

      banks_to_threads[col][m[idx].t] = 1;
    }

    printf ("\n");
  }

  // verify correctness of the pattern
  // each entry has to be unique

  for (int m1 = 0; m1 < CHUNKS; m1++)
  {
    if (m[m1].t == -1) continue; // ignore uninitialized

    for (int m2 = m1 + 1; m2 < CHUNKS; m2++)
    {
      if (m[m2].t == -1) continue; // ignore uninitialized

      if (m[m1].t != m[m2].t) continue;
      if (m[m1].c != m[m2].c) continue;
      if (m[m1].k != m[m2].k) continue;

      printf ("access pattern incorrect\n");

      return -1;
    }
  }

  // visualize bank conflicts
  // the less per line the better

  printf ("thread access at bank :\n");

  for (int col = 0; col < BANKS; col++)
  {
    printf ("%2d: ", col);

    for (int t = 0; t < THREADS; t++)
    {
      if (banks_to_threads[col][t] == 1)
      {
        printf ("%2d ", t);
      }
    }

    printf ("\n");
  }

  return 0;
}
